C******************************************************************************
      
C  Routine:     checkpoint_wr()
      
C  Description: Create a checkpoint file from which the simulation can be
C               restarted.  This version writes unformatted Fortran binary
C               data.  Currently writes are done serially by the master
C               processor using unformatted sequential I/O.
      
C  Authors:     K. Olson and C. Mobarry 7/97
      
      subroutine checkpoint_wr_f77_unf (filenum, simtime)
c
c modified 10-28-99 to split the data across multiple files
c if the data size > 2 GB
c
c This version of the checkpoint routine is identical to the one in FLASH,
c except the facevars are not stored.  This is to enable a direct comparision
c to the HDF routines.

C==============================================================================

#include "common.fh"
      
      integer           filenum
      real              simtime
      
      integer           block_no, jproc, i, j, ivar, iunit1, ix, iy, iz,
     &     iunit2, lnblockst, ngid
      integer           n_to_left(0:16383) ! must extend from 0 to NumPEs-1
      integer           gid(mfaces+1+mchild,maxblocks_tr)
      character*4       fnum_string
      character*80      filename
      
      integer           tot_blocks
      integer           lrefinet(maxblocks_tr), nodetypet(maxblocks_tr)
      integer           gidt(mfaces+1+mchild,maxblocks_tr)
      integer           status(MPI_STATUS_SIZE)

      real              coordt(mdim,maxblocks_tr)
      real              sizet(mdim,maxblocks_tr)
      real              bnd_boxt(2,mdim,maxblocks_tr)
      real          unkt(nvar,il_bnd:iu_bnd,jl_bnd:ju_bnd,kl_bnd:ku_bnd)
      real          facevarxt(nbndvar,il_bnd:iu_bnd+1,jl_bnd:ju_bnd,
     &     kl_bnd:ku_bnd)
      real          facevaryt(nbndvar,il_bnd:iu_bnd,jl_bnd:ju_bnd+k2d,
     &     kl_bnd:ku_bnd)
      real          facevarzt(nbndvar,il_bnd:iu_bnd,jl_bnd:ju_bnd,
     &     kl_bnd:ku_bnd+k3d)
      real              xleft, deltax, xpos
      
      save gid, gidt, lrefinet, nodetypet, lnblockst, coordt,
     &     sizet, bnd_boxt, n_to_left, unkt, facevarxt,
     &     facevaryt, facevarzt

c maximum file size -- if we exceed this, split the data into multiple files
c specify in bytes
      real MAX_FILE_SIZE
      parameter (MAX_FILE_SIZE = 1.8e9)

      integer MAX_FILES
      parameter (MAX_FILES = 9)

      integer num_blocks(MAX_FILES)

      integer jproc_count, ifile_out, num_procs_per_file, num_files
      real total_size
      character file_string*1

      integer ierr, FAIL
      parameter (FAIL = -1)

C=============================================================================

C COMPUTE TOTAL NO. OF BLOCKS STORED TO THE LEFT OF THIS PROCESSOR


      call MPI_BARRIER(MPI_COMM_WORLD,ierr)

      do i = 0,NumPEs-1
         if (i.eq.MyPe) then
            n_to_left(i)=lnblocks
         end if
         call MPI_BCAST(n_to_left(i),1,MPI_INTEGER,i,
     &        MPI_COMM_WORLD,ierr)
      end do
      
      tot_blocks = 0
      do i = 0,NumPEs-1
         tot_blocks = tot_blocks + n_to_left(i)
      end do
      
      do i = NumPEs-1,1,-1
         n_to_left(i) = n_to_left(i-1)
      end do

      n_to_left(0) = 0
      do i = 2,NumPEs-1
         n_to_left(i) = n_to_left(i) + n_to_left(i-1)
      end do

      call MPI_BARRIER(MPI_COMM_WORLD,ierr)

C COMPUTE GLOBAL INDIRECT ADDRESSES FOR TREE DATA (gid)

      do block_no = 1,lnblocks

         ngid = 0
         do j = 1,nfaces
            ngid = ngid + 1
            if (neigh(1,j,block_no).gt.0) then
               gid(ngid,block_no) = neigh(1,j,block_no) + 
     $              n_to_left(neigh(2,j,block_no))
            else
               gid(ngid,block_no) = neigh(1,j,block_no)
            end if
         end do
        
         ngid = ngid + 1
         if (parent(1,block_no).gt.0) then
            gid(ngid,block_no) = parent(1,block_no) + 
     $           n_to_left(parent(2,block_no))
         else
            gid(ngid,block_no) = parent(1,block_no)
         end if
         
         do j = 1,nchild
            ngid = ngid + 1
            if (child(1,j,block_no).gt.0) then
               gid(ngid,block_no) = child(1,j,block_no) + 
     $              n_to_left(child(2,j,block_no))
            else
               gid(ngid,block_no) = child(1,j,block_no)
            end if
         end do

      end do

      call MPI_BARRIER(MPI_COMM_WORLD,ierr)

c we now have the total number of blocks, compute the number of files
c we need to store this in
      total_size = real(nvar*nxb*nyb*nzb*tot_blocks)*8.
      num_files = int(total_size/MAX_FILE_SIZE) + 1

      if (mype.eq.0) then
         open (unit=30,file='amr_log',status='unknown',
     &        position='append',form='formatted')
         write (30,*) ' num_files = ',num_files
         write (30,*) ' total_size = ',total_size
         write (30,*) ' tot_blocks = ',tot_blocks
         write (30,*) ' nvar,nxb,nyb,nzb = ',nvar,nxb,nyb,nzb
         write (30,*) ' MAX_FILE_SIZE = ',MAX_FILE_SIZE
         close(30)
      end if

      if (num_files .GT. NumPEs) then
         print *, 'Error: to split data across multiple files, there '
         print *, 'must be atleast 1 processor / file'
         print *, ' '
         print *, 'number of files = ', num_files
         call MPI_Abort(MPI_COMM_WORLD, FAIL, ierr)
      endif

c compute the number of processors / file (assuming they all have the same
c number of blocks)
      num_procs_per_file = NumPEs/num_files

c if there are more than 1 file, then figure out how many blocks it will
c store.  This number will be used to dimension the hdf unk array and 
c other arrays that need the number of blocks as a dimension
      if (num_files .EQ. 1) then
         num_blocks(1) = tot_blocks
      else
         do i = 1, num_files-1
            num_blocks(i) = n_to_left(i*num_procs_per_file) -
     +           n_to_left((i-1)*num_procs_per_file)
         enddo
         
         num_blocks(num_files) = tot_blocks - 
     +        n_to_left((num_files-1)*num_procs_per_file)

         if (MyPE .EQ. MasterPE) then 
            print *, ' '
            print *, '*** using ', num_files, ' files to store data ***'
         endif

      endif

      call MPI_BARRIER(MPI_COMM_WORLD,ierr)


C NOW WRITE OUT THE DATA FROM PROC 0

      if (MyPe .eq. 0) then

c create the filename -- use the basename and file #         
         write (fnum_string, '(i4.4)') filenum
         filename = trim(basenm) // 'f77_chk_'// fnum_string
         iunit1 = 10

c keep track of what file we are writing out to
         ifile_out = 1

         open (file=filename, unit=iunit1, form='unformatted',
     &        status='unknown')

c the first record now includes the local number of blocks
c note, this makes these files incompatible with older files
c         write (iunit1) tot_blocks,simtime,dt
         write (iunit1) tot_blocks,simtime,dt, num_blocks(ifile_out)
      end if

c keep track of the number of processors written out to the file
      jproc_count = 0

c-----------------------------------------------------------------------------
c begin the loop over processors
c-----------------------------------------------------------------------------
      do jproc = 0,NumPEs-1

         jproc_count = jproc_count + 1

         if (MyPe.eq.MasterPE) then

c determine if we need to open the next file
            if (jproc_count .GT. num_procs_per_file) then

c reset the processor and block count
               jproc_count = 1
               
c increment the file counter
               ifile_out = ifile_out + 1

c close the current file
               close (iunit1)
               print *, '*** Wrote output to ', trim(filename), 
     +              ' (', num_blocks(ifile_out), 'blocks ) ***'

c open the next file
               write (file_string, '(i1.1)') ifile_out

               filename = trim(basenm) // 'f77_chk_' // fnum_string //
     +              '_' // trim(file_string)

               open (file=filename, unit=iunit1, form='unformatted',
     &              status='unknown')

c write out the header information
               write (iunit1) tot_blocks,simtime,dt, 
     +              num_blocks(ifile_out)

            endif

C fetch lnblocks from other processors

            if (jproc.ne.0) then

            call MPI_RECV (lnblockst,1,MPI_INTEGER,jproc,
     &              1,MPI_COMM_WORLD,status,ierr)
            if (lnblockst.gt.0) then
               call MPI_RECV (lrefinet(1),lnblockst,MPI_INTEGER,jproc,
     &              2,MPI_COMM_WORLD,status,ierr)
               call MPI_RECV (nodetypet(1),lnblockst,MPI_INTEGER,jproc,
     &              3,MPI_COMM_WORLD,status,ierr)
               call MPI_RECV (coordt(1,1),mdim*lnblockst,
     &              MPI_DOUBLE_PRECISION,
     &              jproc,4,MPI_COMM_WORLD,status,ierr)
               call MPI_RECV (sizet(1,1),mdim*lnblockst,
     &              MPI_DOUBLE_PRECISION,
     &              jproc,5,MPI_COMM_WORLD,status,ierr)
               call MPI_RECV (bnd_boxt(1,1,1),2*mdim*lnblockst,
     &              MPI_DOUBLE_PRECISION,
     &              jproc,6,MPI_COMM_WORLD,status,ierr)
               call MPI_RECV (gidt(1,1),lnblockst*(mfaces+1+mchild),
     &              MPI_INTEGER,
     &              jproc,7,MPI_COMM_WORLD,status,ierr)
            end if

            else
               
               lnblockst = lnblocks
               lrefinet(1:lnblockst)  = lrefine(1:lnblockst)
               nodetypet(1:lnblockst) = nodetype(1:lnblockst)
               coordt(:,1:lnblockst) = coord(:,1:lnblockst)
               sizet(:,1:lnblockst) = size(:,1:lnblockst)
               bnd_boxt(:,:,1:lnblockst) = bnd_box(:,:,1:lnblockst)
               gidt(:,1:lnblockst) = gid(:,1:lnblockst)

            end if

            do block_no = 1,lnblockst

C fetch data for this block

               if (jproc.ne.0) then
                  call MPI_RECV (unkt(1,1,1,1),
     &                 len_block,
     &                 MPI_DOUBLE_PRECISION,
     &                 jproc,7+block_no,
     &                 MPI_COMM_WORLD,status,ierr)
                  if (nfacevar.gt.0) then
                  call MPI_RECV (facevarxt(1,1,1,1),
     &                 nbndvar*len_blockfx,
     &                 MPI_DOUBLE_PRECISION,
     &                 jproc,7+block_no+lnblockst,
     &                 MPI_COMM_WORLD,status,ierr)
                  call MPI_RECV (facevaryt(1,1,1,1),
     &                 nbndvar*len_blockfy,
     &                 MPI_DOUBLE_PRECISION,
     &                 jproc,7+block_no+2*lnblockst,
     &                 MPI_COMM_WORLD,status,ierr)
                  call MPI_RECV (facevarzt(1,1,1,1),
     &                 nbndvar*len_blockfz,
     &                 MPI_DOUBLE_PRECISION,
     &                 jproc,7+block_no+3*lnblockst,
     &                 MPI_COMM_WORLD,status,ierr)
                  end if
               else
                  unkt(:,:,:,:) = unk(:,:,:,:,block_no)
                  if (nfacevar.gt.0) then
                     facevarxt(:,:,:,:) = facevarx(:,:,:,:,block_no)
                     facevaryt(:,:,:,:) = facevary(:,:,:,:,block_no)
                     facevarzt(:,:,:,:) = facevarz(:,:,:,:,block_no)
                  end if
               end if

c               write (iunit1) 
c     &              lrefinet(block_no),
c     &              nodetypet(block_no),
c     &              (gidt(j,block_no),j=1,nfaces+1+nchild),
c     &              (coordt(j,block_no),j=1,ndim),
c     &              (sizet(j,block_no),j=1,ndim),
c     &              (bnd_boxt(1,j,block_no),j=1,ndim),
c     &              (bnd_boxt(2,j,block_no),j=1,ndim),
c     &              ((((unkt(ivar,ix,iy,iz),
c     &                 ivar = 1,nvar),
c     &                 ix = 1+nguard,nxb+nguard),
c     &                 iy = 1+nguard,nyb+nguard),
c     &                 iz = 1+nguard*k3d,nzb+nguard*k3d),
c     &              ((((facevarxt(ivar,ix,iy,iz),
c     &                 ivar = 1,nbndvar),
c     &                 ix = 1+nguard,nxb+nguard+1),
c     &                 iy = 1+nguard,nyb+nguard),
c     &                 iz = 1+nguard*k3d,nzb+nguard*k3d),
c     &              ((((facevaryt(ivar,ix,iy,iz),
c     &                 ivar = 1,nbndvar),
c     &                 ix = 1+nguard,nxb+nguard),
c     &                 iy = 1+nguard,nyb+nguard+1),
c     &                 iz = 1+nguard*k3d,nzb+nguard*k3d),
c     &              ((((facevarzt(ivar,ix,iy,iz),
c     &                 ivar = 1,nbndvar),
c     &                 ix = 1+nguard,nxb+nguard),
c     &                 iy = 1+nguard,nyb+nguard),
c     &                 iz = 1+nguard*k3d,nzb+(nguard+1)*k3d)

               write (iunit1) 
     &              lrefinet(block_no),
     &              nodetypet(block_no),
     &              (gidt(j,block_no),j=1,nfaces+1+nchild),
     &              (coordt(j,block_no),j=1,ndim),
     &              (sizet(j,block_no),j=1,ndim),
     &              (bnd_boxt(1,j,block_no),j=1,ndim),
     &              (bnd_boxt(2,j,block_no),j=1,ndim),
     &              ((((unkt(ivar,ix,iy,iz),
     &                 ivar = 1,nvar),
     &                 ix = 1+nguard,nxb+nguard),
     &                 iy = 1+nguard,nyb+nguard),
     &                 iz = 1+nguard*k3d,nzb+nguard*k3d)

            end do

         else ! if (MyPe.eq.MasterPE

            if (jproc.eq.MyPe) then

               call MPI_SEND (lnblocks,1,MPI_INTEGER,0,
     &              1,MPI_COMM_WORLD,ierr)
               if (lnblocks.gt.0) then
                  call MPI_SEND (lrefine(1),lnblocks,MPI_INTEGER,0,
     &                 2,MPI_COMM_WORLD,ierr)
                  call MPI_SEND (nodetype(1),lnblocks,MPI_INTEGER,0,
     &                 3,MPI_COMM_WORLD,ierr)
                  call MPI_SEND (coord(1,1),mdim*lnblocks,
     &                 MPI_DOUBLE_PRECISION,0,
     &                 4,MPI_COMM_WORLD,ierr)
                  call MPI_SEND (size(1,1),mdim*lnblocks,
     &                 MPI_DOUBLE_PRECISION,0,
     &                 5,MPI_COMM_WORLD,ierr)
                  call MPI_SEND (bnd_box(1,1,1),2*mdim*lnblocks,
     &                 MPI_DOUBLE_PRECISION,
     &                 0,6,MPI_COMM_WORLD,ierr)
                  call MPI_SEND (gid(1,1),lnblocks*(mfaces+1+mchild),
     &                 MPI_INTEGER,
     &                 0,7,MPI_COMM_WORLD,ierr)
                  do block_no = 1,lnblocks
                     call MPI_SEND (unk(1,1,1,1,block_no),
     &                    len_block,
     &                    MPI_DOUBLE_PRECISION,
     &                    0,7+block_no,
     &                    MPI_COMM_WORLD,ierr)
                     if (nfacevar.gt.0) then
                        call MPI_SEND (facevarx(1,1,1,1,block_no),
     &                       nbndvar*len_blockfx,
     &                       MPI_DOUBLE_PRECISION,
     &                       0,7+block_no+lnblocks,
     &                       MPI_COMM_WORLD,ierr)
                        call MPI_SEND (facevary(1,1,1,1,block_no),
     &                       nbndvar*len_blockfy,
     &                       MPI_DOUBLE_PRECISION,
     &                       0,7+block_no+2*lnblocks,
     &                       MPI_COMM_WORLD,ierr)
                        call MPI_SEND (facevarz(1,1,1,1,block_no),
     &                       nbndvar*len_blockfz,
     &                       MPI_DOUBLE_PRECISION,
     &                       0,7+block_no+3*lnblocks,
     &                       MPI_COMM_WORLD,ierr)
                     end if
                  end do
               end if
               
            end if

         end if                 ! if MyPe == 0
      
         call MPI_BARRIER(MPI_COMM_WORLD,ierr)

      end do  ! end loop over processors

      if (MyPe.eq.MasterPE) then
         close(iunit1)
         print *, '*** Wrote output to ', trim(filename), 
     +        ' (', num_blocks(ifile_out), 'blocks ) ***'
         if (num_files .GT. 1) print *, ' '
      end if

      call MPI_BARRIER(MPI_COMM_WORLD,ierr)

      return
      end

