      program flash_benchmark_io
c
c This is a sample program that setups the FLASH data structures and 
c drives the I/O routines.  It is intended for benchmarking the I/O
c performance.
c 

c the main data structures are contained in common blocks, defined in the
c include files
#include "common.fh"

      integer ierr
      integer i

      real time_io, time_begin

      integer, parameter :: local_blocks = 0.8*maxblocks

c initialize MPI and get the rank and size
      call MPI_INIT(ierr)
      
      call MPI_Comm_Rank (MPI_Comm_World, MyPE, ierr)
      call MPI_Comm_Size (MPI_Comm_World, NumPEs, ierr)

      MasterPE = 0

      call init_materials

      if (MyPE .EQ. MasterPE) then

c print some initialization information
         print *, 'HDF 5 v 1.4 test on Blue Pacific'
         print *, NumPEs, ' processors'
         print *, 'nxb, nyb, nzb = ', nxb, nyb, nzb
         print *, 'nguard = ', nguard
         print *, 'number of blocks = ', local_blocks
         print *, 'nvar = ', nvar
      endif


c put ~100 blocks on each processor -- make it vary a little, since it does
c in the real application.  This is the maximum that we can fit on Blue 
c Pacific comfortably.
      lnblocks = local_blocks + mod(MyPE,3)

c just fill the tree stucture with dummy information -- we are just going to
c dump it out
      size(:,:) = 0.5e0
      lrefine(:) = 1
      nodetype(:) = 1
      refine(:) = .FALSE.
      derefine(:) = .FALSE.
      parent(:,:) = -1
      child(:,:,:) = -1
      coord(:,:) = 0.25e0
      bnd_box(:,:,:) = 0.e0
      neigh(:,:,:) = -1
      empty(:) = 0

c initialize the unknowns with the index of the variable
      do i = 1, nvar
        unk(i,:,:,:,:) = float(i)
      enddo

      facevarx(:,:,:,:,:) = 0.
      facevary(:,:,:,:,:) = 0.
      facevarz(:,:,:,:,:) = 0.

c setup the file properties
      basenm = "flash_io_test_"

#ifdef CHIBA
      basenm = "pvfs:/stopvfsvol/flash_io_test_"
#endif      


c write out a f77 checkpoint -- collectively
c      if (MyPE .EQ. MasterPE) then
c         print *, ' '
c         print *, 'dumping out a single f77 file from proc 0'
c         print *, 'data from all procs is moved to proc 0'
c      endif
c      
c      time_begin = MPI_Wtime()
c      call checkpoint_wr_f77_unf(0,0.e0)
c      time_io = MPI_Wtime() - time_begin
c
c      if (MyPE .EQ. MasterPE) then
c         print *, 'time to output = ', time_io
c      endif

c---------------------------------------------------------------------------
c HDF 5 checkpoint file
c---------------------------------------------------------------------------
      if (MyPE .EQ. MasterPE) then
         print *, ' '
         print *, 'doing HDF 5 parallel I/O to a single file'
      endif

      time_begin = MPI_Wtime()
      call checkpoint_wr_hdf5_par(0,0.e0)
      time_io = MPI_Wtime() - time_begin

      if (MyPE .EQ. MasterPE) then
         print *, 'time to output = ', time_io
      endif

c---------------------------------------------------------------------------
c HDF 5 plotfile -- no corners
c---------------------------------------------------------------------------
      if (MyPE .EQ. MasterPE) then
         print *, ' '
         print *, 'writing an HDF 5 plotfile --  no corners'
      endif

      time_begin = MPI_Wtime()
      call plotfile_hdf5_par(0,0.e0,.false.)
      time_io = MPI_Wtime() - time_begin
    
      if (MyPE .EQ. MasterPE) then
         print *, 'time to output = ', time_io
      endif

c---------------------------------------------------------------------------
c HDF 5 plotfile -- corners
c---------------------------------------------------------------------------
      if (MyPE .EQ. MasterPE) then
         print *, ' '
         print *, 'writing an HDF 5 plotfile --  corners'
      endif

      time_begin = MPI_Wtime()
      call plotfile_hdf5_par(0,0.e0,.true.)
      time_io = MPI_Wtime() - time_begin
    
      if (MyPE .EQ. MasterPE) then
         print *, 'time to output = ', time_io
      endif


      call MPI_Finalize(ierr)
      end





